# DO NOT EDIT! This file is automatically generated
import datetime
import enum
import typing

from commercetools.types._abstract import _BaseType

if typing.TYPE_CHECKING:
    from ._channel import ChannelReference, ChannelResourceIdentifier
    from ._customer import CustomerReference
    from ._customer_group import CustomerGroupReference, CustomerGroupResourceIdentifier
    from ._product import FacetResults
    from ._product_discount import ProductDiscountReference
    from ._type import CustomFields, CustomFieldsDraft
__all__ = [
    "Address",
    "Asset",
    "AssetDimensions",
    "AssetDraft",
    "AssetSource",
    "BaseResource",
    "CentPrecisionMoney",
    "CentPrecisionMoneyDraft",
    "ClientLogging",
    "CreatedBy",
    "DiscountedPrice",
    "GeoJson",
    "GeoJsonPoint",
    "HighPrecisionMoney",
    "HighPrecisionMoneyDraft",
    "Image",
    "ImageDimensions",
    "KeyReference",
    "LastModifiedBy",
    "LocalizedString",
    "Money",
    "MoneyType",
    "PagedQueryResponse",
    "Price",
    "PriceDraft",
    "PriceTier",
    "PriceTierDraft",
    "QueryPrice",
    "Reference",
    "ReferenceTypeId",
    "ResourceIdentifier",
    "ScopedPrice",
    "TypedMoney",
    "TypedMoneyDraft",
    "Update",
    "UpdateAction",
]


class Address(_BaseType):
    #: Optional :class:`str`
    id: typing.Optional[str]
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`str`
    title: typing.Optional[str]
    #: Optional :class:`str`
    salutation: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``firstName`` `in Commercetools)`
    first_name: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``lastName`` `in Commercetools)`
    last_name: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``streetName`` `in Commercetools)`
    street_name: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``streetNumber`` `in Commercetools)`
    street_number: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``additionalStreetInfo`` `in Commercetools)`
    additional_street_info: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``postalCode`` `in Commercetools)`
    postal_code: typing.Optional[str]
    #: Optional :class:`str`
    city: typing.Optional[str]
    #: Optional :class:`str`
    region: typing.Optional[str]
    #: Optional :class:`str`
    state: typing.Optional[str]
    #: :class:`str`
    country: "str"
    #: Optional :class:`str`
    company: typing.Optional[str]
    #: Optional :class:`str`
    department: typing.Optional[str]
    #: Optional :class:`str`
    building: typing.Optional[str]
    #: Optional :class:`str`
    apartment: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``pOBox`` `in Commercetools)`
    p_o_box: typing.Optional[str]
    #: Optional :class:`str`
    phone: typing.Optional[str]
    #: Optional :class:`str`
    mobile: typing.Optional[str]
    #: Optional :class:`str`
    email: typing.Optional[str]
    #: Optional :class:`str`
    fax: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``additionalAddressInfo`` `in Commercetools)`
    additional_address_info: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``externalId`` `in Commercetools)`
    external_id: typing.Optional[str]

    def __init__(
        self,
        *,
        country: "str",
        id: typing.Optional[str] = None,
        key: typing.Optional[str] = None,
        title: typing.Optional[str] = None,
        salutation: typing.Optional[str] = None,
        first_name: typing.Optional[str] = None,
        last_name: typing.Optional[str] = None,
        street_name: typing.Optional[str] = None,
        street_number: typing.Optional[str] = None,
        additional_street_info: typing.Optional[str] = None,
        postal_code: typing.Optional[str] = None,
        city: typing.Optional[str] = None,
        region: typing.Optional[str] = None,
        state: typing.Optional[str] = None,
        company: typing.Optional[str] = None,
        department: typing.Optional[str] = None,
        building: typing.Optional[str] = None,
        apartment: typing.Optional[str] = None,
        p_o_box: typing.Optional[str] = None,
        phone: typing.Optional[str] = None,
        mobile: typing.Optional[str] = None,
        email: typing.Optional[str] = None,
        fax: typing.Optional[str] = None,
        additional_address_info: typing.Optional[str] = None,
        external_id: typing.Optional[str] = None,
    ) -> None:
        self.id = id
        self.key = key
        self.title = title
        self.salutation = salutation
        self.first_name = first_name
        self.last_name = last_name
        self.street_name = street_name
        self.street_number = street_number
        self.additional_street_info = additional_street_info
        self.postal_code = postal_code
        self.city = city
        self.region = region
        self.state = state
        self.country = country
        self.company = company
        self.department = department
        self.building = building
        self.apartment = apartment
        self.p_o_box = p_o_box
        self.phone = phone
        self.mobile = mobile
        self.email = email
        self.fax = fax
        self.additional_address_info = additional_address_info
        self.external_id = external_id
        super().__init__()

    def __repr__(self) -> str:
        return (
            "Address(id=%r, key=%r, title=%r, salutation=%r, first_name=%r, last_name=%r, street_name=%r, street_number=%r, additional_street_info=%r, postal_code=%r, city=%r, region=%r, state=%r, country=%r, company=%r, department=%r, building=%r, apartment=%r, p_o_box=%r, phone=%r, mobile=%r, email=%r, fax=%r, additional_address_info=%r, external_id=%r)"
            % (
                self.id,
                self.key,
                self.title,
                self.salutation,
                self.first_name,
                self.last_name,
                self.street_name,
                self.street_number,
                self.additional_street_info,
                self.postal_code,
                self.city,
                self.region,
                self.state,
                self.country,
                self.company,
                self.department,
                self.building,
                self.apartment,
                self.p_o_box,
                self.phone,
                self.mobile,
                self.email,
                self.fax,
                self.additional_address_info,
                self.external_id,
            )
        )


class Asset(_BaseType):
    #: :class:`str`
    id: str
    #: List of :class:`commercetools.types.AssetSource`
    sources: typing.List["AssetSource"]
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: Optional list of :class:`str`
    tags: typing.Optional[typing.List[str]]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        id: str,
        sources: typing.List["AssetSource"],
        name: "LocalizedString",
        description: typing.Optional["LocalizedString"] = None,
        tags: typing.Optional[typing.List[str]] = None,
        custom: typing.Optional["CustomFields"] = None,
        key: typing.Optional[str] = None,
    ) -> None:
        self.id = id
        self.sources = sources
        self.name = name
        self.description = description
        self.tags = tags
        self.custom = custom
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return (
            "Asset(id=%r, sources=%r, name=%r, description=%r, tags=%r, custom=%r, key=%r)"
            % (
                self.id,
                self.sources,
                self.name,
                self.description,
                self.tags,
                self.custom,
                self.key,
            )
        )


class AssetDimensions(_BaseType):
    #: :class:`int`
    w: int
    #: :class:`int`
    h: int

    def __init__(self, *, w: int, h: int) -> None:
        self.w = w
        self.h = h
        super().__init__()

    def __repr__(self) -> str:
        return "AssetDimensions(w=%r, h=%r)" % (self.w, self.h)


class AssetDraft(_BaseType):
    #: List of :class:`commercetools.types.AssetSource`
    sources: typing.List["AssetSource"]
    #: :class:`commercetools.types.LocalizedString`
    name: "LocalizedString"
    #: Optional :class:`commercetools.types.LocalizedString`
    description: typing.Optional["LocalizedString"]
    #: Optional list of :class:`str`
    tags: typing.Optional[typing.List[str]]
    #: Optional :class:`commercetools.types.CustomFieldsDraft`
    custom: typing.Optional["CustomFieldsDraft"]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        sources: typing.List["AssetSource"],
        name: "LocalizedString",
        description: typing.Optional["LocalizedString"] = None,
        tags: typing.Optional[typing.List[str]] = None,
        custom: typing.Optional["CustomFieldsDraft"] = None,
        key: typing.Optional[str] = None,
    ) -> None:
        self.sources = sources
        self.name = name
        self.description = description
        self.tags = tags
        self.custom = custom
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return (
            "AssetDraft(sources=%r, name=%r, description=%r, tags=%r, custom=%r, key=%r)"
            % (
                self.sources,
                self.name,
                self.description,
                self.tags,
                self.custom,
                self.key,
            )
        )


class AssetSource(_BaseType):
    #: :class:`str`
    uri: str
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: Optional :class:`commercetools.types.AssetDimensions`
    dimensions: typing.Optional["AssetDimensions"]
    #: Optional :class:`str` `(Named` ``contentType`` `in Commercetools)`
    content_type: typing.Optional[str]

    def __init__(
        self,
        *,
        uri: str,
        key: typing.Optional[str] = None,
        dimensions: typing.Optional["AssetDimensions"] = None,
        content_type: typing.Optional[str] = None,
    ) -> None:
        self.uri = uri
        self.key = key
        self.dimensions = dimensions
        self.content_type = content_type
        super().__init__()

    def __repr__(self) -> str:
        return "AssetSource(uri=%r, key=%r, dimensions=%r, content_type=%r)" % (
            self.uri,
            self.key,
            self.dimensions,
            self.content_type,
        )


class BaseResource(_BaseType):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        super().__init__()

    def __repr__(self) -> str:
        return "BaseResource(id=%r, version=%r, created_at=%r, last_modified_at=%r)" % (
            self.id,
            self.version,
            self.created_at,
            self.last_modified_at,
        )


class ClientLogging(_BaseType):
    #: Optional :class:`str` `(Named` ``clientId`` `in Commercetools)`
    client_id: typing.Optional[str]
    #: Optional :class:`str` `(Named` ``externalUserId`` `in Commercetools)`
    external_user_id: typing.Optional[str]
    #: Optional :class:`commercetools.types.CustomerReference`
    customer: typing.Optional["CustomerReference"]
    #: Optional :class:`str` `(Named` ``anonymousId`` `in Commercetools)`
    anonymous_id: typing.Optional[str]

    def __init__(
        self,
        *,
        client_id: typing.Optional[str] = None,
        external_user_id: typing.Optional[str] = None,
        customer: typing.Optional["CustomerReference"] = None,
        anonymous_id: typing.Optional[str] = None,
    ) -> None:
        self.client_id = client_id
        self.external_user_id = external_user_id
        self.customer = customer
        self.anonymous_id = anonymous_id
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ClientLogging(client_id=%r, external_user_id=%r, customer=%r, anonymous_id=%r)"
            % (self.client_id, self.external_user_id, self.customer, self.anonymous_id)
        )


class DiscountedPrice(_BaseType):
    #: :class:`commercetools.types.Money`
    value: "Money"
    #: :class:`commercetools.types.ProductDiscountReference`
    discount: "ProductDiscountReference"

    def __init__(self, *, value: "Money", discount: "ProductDiscountReference") -> None:
        self.value = value
        self.discount = discount
        super().__init__()

    def __repr__(self) -> str:
        return "DiscountedPrice(value=%r, discount=%r)" % (self.value, self.discount)


class GeoJson(_BaseType):
    #: :class:`str`
    type: str

    def __init__(self, *, type: str) -> None:
        self.type = type
        super().__init__()

    def __repr__(self) -> str:
        return "GeoJson(type=%r)" % (self.type,)


class Image(_BaseType):
    #: :class:`str`
    url: str
    #: :class:`commercetools.types.ImageDimensions`
    dimensions: "ImageDimensions"
    #: Optional :class:`str`
    label: typing.Optional[str]

    def __init__(
        self,
        *,
        url: str,
        dimensions: "ImageDimensions",
        label: typing.Optional[str] = None,
    ) -> None:
        self.url = url
        self.dimensions = dimensions
        self.label = label
        super().__init__()

    def __repr__(self) -> str:
        return "Image(url=%r, dimensions=%r, label=%r)" % (
            self.url,
            self.dimensions,
            self.label,
        )


class ImageDimensions(_BaseType):
    #: :class:`int`
    w: int
    #: :class:`int`
    h: int

    def __init__(self, *, w: int, h: int) -> None:
        self.w = w
        self.h = h
        super().__init__()

    def __repr__(self) -> str:
        return "ImageDimensions(w=%r, h=%r)" % (self.w, self.h)


class KeyReference(_BaseType):
    #: :class:`commercetools.types.ReferenceTypeId` `(Named` ``typeId`` `in Commercetools)`
    type_id: "ReferenceTypeId"
    #: :class:`str`
    key: str

    def __init__(self, *, type_id: "ReferenceTypeId", key: str) -> None:
        self.type_id = type_id
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return "KeyReference(type_id=%r, key=%r)" % (self.type_id, self.key)


class LocalizedString(typing.Dict[str, str]):
    def __repr__(self) -> str:
        return "LocalizedString(%s)" % (
            ", ".join(f"{k}={v!r}" for k, v in self.items())
        )


class Money(_BaseType):
    #: :class:`int` `(Named` ``centAmount`` `in Commercetools)`
    cent_amount: int
    #: :class:`str` `(Named` ``currencyCode`` `in Commercetools)`
    currency_code: "str"

    def __init__(self, *, cent_amount: int, currency_code: "str") -> None:
        self.cent_amount = cent_amount
        self.currency_code = currency_code
        super().__init__()

    def __repr__(self) -> str:
        return "Money(cent_amount=%r, currency_code=%r)" % (
            self.cent_amount,
            self.currency_code,
        )


class MoneyType(enum.Enum):
    CENT_PRECISION = "centPrecision"
    HIGH_PRECISION = "highPrecision"


class PagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.BaseResource`
    results: typing.Sequence["BaseResource"]
    #: Optional :class:`commercetools.types.FacetResults`
    facets: typing.Optional["FacetResults"]
    #: Optional :class:`object`
    meta: typing.Optional[object]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["BaseResource"],
        total: typing.Optional[int] = None,
        facets: typing.Optional["FacetResults"] = None,
        meta: typing.Optional[object] = None,
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        self.facets = facets
        self.meta = meta
        super().__init__()

    def __repr__(self) -> str:
        return (
            "PagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r, facets=%r, meta=%r)"
            % (
                self.limit,
                self.count,
                self.total,
                self.offset,
                self.results,
                self.facets,
                self.meta,
            )
        )


class Price(_BaseType):
    #: :class:`str`
    id: str
    #: :class:`commercetools.types.TypedMoney`
    value: "TypedMoney"
    #: Optional :class:`str`
    country: typing.Optional["str"]
    #: Optional :class:`commercetools.types.CustomerGroupReference` `(Named` ``customerGroup`` `in Commercetools)`
    customer_group: typing.Optional["CustomerGroupReference"]
    #: Optional :class:`commercetools.types.ChannelReference`
    channel: typing.Optional["ChannelReference"]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.DiscountedPrice`
    discounted: typing.Optional["DiscountedPrice"]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]
    #: Optional list of :class:`commercetools.types.PriceTier`
    tiers: typing.Optional[typing.List["PriceTier"]]

    def __init__(
        self,
        *,
        id: str,
        value: "TypedMoney",
        country: typing.Optional["str"] = None,
        customer_group: typing.Optional["CustomerGroupReference"] = None,
        channel: typing.Optional["ChannelReference"] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        discounted: typing.Optional["DiscountedPrice"] = None,
        custom: typing.Optional["CustomFields"] = None,
        tiers: typing.Optional[typing.List["PriceTier"]] = None,
    ) -> None:
        self.id = id
        self.value = value
        self.country = country
        self.customer_group = customer_group
        self.channel = channel
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.discounted = discounted
        self.custom = custom
        self.tiers = tiers
        super().__init__()

    def __repr__(self) -> str:
        return (
            "Price(id=%r, value=%r, country=%r, customer_group=%r, channel=%r, valid_from=%r, valid_until=%r, discounted=%r, custom=%r, tiers=%r)"
            % (
                self.id,
                self.value,
                self.country,
                self.customer_group,
                self.channel,
                self.valid_from,
                self.valid_until,
                self.discounted,
                self.custom,
                self.tiers,
            )
        )


class PriceDraft(_BaseType):
    #: :class:`commercetools.types.Money`
    value: "Money"
    #: Optional :class:`str`
    country: typing.Optional["str"]
    #: Optional :class:`commercetools.types.CustomerGroupResourceIdentifier` `(Named` ``customerGroup`` `in Commercetools)`
    customer_group: typing.Optional["CustomerGroupResourceIdentifier"]
    #: Optional :class:`commercetools.types.ChannelResourceIdentifier`
    channel: typing.Optional["ChannelResourceIdentifier"]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.CustomFieldsDraft`
    custom: typing.Optional["CustomFieldsDraft"]
    #: Optional list of :class:`commercetools.types.PriceTierDraft`
    tiers: typing.Optional[typing.List["PriceTierDraft"]]
    #: Optional :class:`commercetools.types.DiscountedPrice`
    discounted: typing.Optional["DiscountedPrice"]

    def __init__(
        self,
        *,
        value: "Money",
        country: typing.Optional["str"] = None,
        customer_group: typing.Optional["CustomerGroupResourceIdentifier"] = None,
        channel: typing.Optional["ChannelResourceIdentifier"] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        custom: typing.Optional["CustomFieldsDraft"] = None,
        tiers: typing.Optional[typing.List["PriceTierDraft"]] = None,
        discounted: typing.Optional["DiscountedPrice"] = None,
    ) -> None:
        self.value = value
        self.country = country
        self.customer_group = customer_group
        self.channel = channel
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.custom = custom
        self.tiers = tiers
        self.discounted = discounted
        super().__init__()

    def __repr__(self) -> str:
        return (
            "PriceDraft(value=%r, country=%r, customer_group=%r, channel=%r, valid_from=%r, valid_until=%r, custom=%r, tiers=%r, discounted=%r)"
            % (
                self.value,
                self.country,
                self.customer_group,
                self.channel,
                self.valid_from,
                self.valid_until,
                self.custom,
                self.tiers,
                self.discounted,
            )
        )


class PriceTier(_BaseType):
    #: :class:`int` `(Named` ``minimumQuantity`` `in Commercetools)`
    minimum_quantity: int
    #: :class:`commercetools.types.TypedMoney`
    value: "TypedMoney"

    def __init__(self, *, minimum_quantity: int, value: "TypedMoney") -> None:
        self.minimum_quantity = minimum_quantity
        self.value = value
        super().__init__()

    def __repr__(self) -> str:
        return "PriceTier(minimum_quantity=%r, value=%r)" % (
            self.minimum_quantity,
            self.value,
        )


class PriceTierDraft(_BaseType):
    #: :class:`int` `(Named` ``minimumQuantity`` `in Commercetools)`
    minimum_quantity: int
    #: :class:`commercetools.types.Money`
    value: "Money"

    def __init__(self, *, minimum_quantity: int, value: "Money") -> None:
        self.minimum_quantity = minimum_quantity
        self.value = value
        super().__init__()

    def __repr__(self) -> str:
        return "PriceTierDraft(minimum_quantity=%r, value=%r)" % (
            self.minimum_quantity,
            self.value,
        )


class QueryPrice(_BaseType):
    #: :class:`str`
    id: str
    #: :class:`commercetools.types.Money`
    value: "Money"
    #: Optional :class:`str`
    country: typing.Optional["str"]
    #: Optional :class:`commercetools.types.CustomerGroupReference` `(Named` ``customerGroup`` `in Commercetools)`
    customer_group: typing.Optional["CustomerGroupReference"]
    #: Optional :class:`commercetools.types.ChannelReference`
    channel: typing.Optional["ChannelReference"]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.DiscountedPrice`
    discounted: typing.Optional["DiscountedPrice"]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]
    #: Optional list of :class:`commercetools.types.PriceTierDraft`
    tiers: typing.Optional[typing.List["PriceTierDraft"]]

    def __init__(
        self,
        *,
        id: str,
        value: "Money",
        country: typing.Optional["str"] = None,
        customer_group: typing.Optional["CustomerGroupReference"] = None,
        channel: typing.Optional["ChannelReference"] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        discounted: typing.Optional["DiscountedPrice"] = None,
        custom: typing.Optional["CustomFields"] = None,
        tiers: typing.Optional[typing.List["PriceTierDraft"]] = None,
    ) -> None:
        self.id = id
        self.value = value
        self.country = country
        self.customer_group = customer_group
        self.channel = channel
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.discounted = discounted
        self.custom = custom
        self.tiers = tiers
        super().__init__()

    def __repr__(self) -> str:
        return (
            "QueryPrice(id=%r, value=%r, country=%r, customer_group=%r, channel=%r, valid_from=%r, valid_until=%r, discounted=%r, custom=%r, tiers=%r)"
            % (
                self.id,
                self.value,
                self.country,
                self.customer_group,
                self.channel,
                self.valid_from,
                self.valid_until,
                self.discounted,
                self.custom,
                self.tiers,
            )
        )


class Reference(_BaseType):
    #: :class:`commercetools.types.ReferenceTypeId` `(Named` ``typeId`` `in Commercetools)`
    type_id: "ReferenceTypeId"
    #: :class:`str`
    id: str

    def __init__(self, *, type_id: "ReferenceTypeId", id: str) -> None:
        self.type_id = type_id
        self.id = id
        super().__init__()

    def __repr__(self) -> str:
        return "Reference(type_id=%r, id=%r)" % (self.type_id, self.id)


class ReferenceTypeId(enum.Enum):
    CART = "cart"
    CART_DISCOUNT = "cart-discount"
    CATEGORY = "category"
    CHANNEL = "channel"
    CUSTOMER = "customer"
    CUSTOMER_GROUP = "customer-group"
    DISCOUNT_CODE = "discount-code"
    KEY_VALUE_DOCUMENT = "key-value-document"
    PAYMENT = "payment"
    PRODUCT = "product"
    PRODUCT_TYPE = "product-type"
    PRODUCT_DISCOUNT = "product-discount"
    ORDER = "order"
    REVIEW = "review"
    SHOPPING_LIST = "shopping-list"
    SHIPPING_METHOD = "shipping-method"
    STATE = "state"
    STORE = "store"
    TAX_CATEGORY = "tax-category"
    TYPE = "type"
    ZONE = "zone"
    INVENTORY_ENTRY = "inventory-entry"
    ORDER_EDIT = "order-edit"


class ResourceIdentifier(_BaseType):
    #: Optional :class:`commercetools.types.ReferenceTypeId` `(Named` ``typeId`` `in Commercetools)`
    type_id: typing.Optional["ReferenceTypeId"]
    #: Optional :class:`str`
    id: typing.Optional[str]
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(
        self,
        *,
        type_id: typing.Optional["ReferenceTypeId"] = None,
        id: typing.Optional[str] = None,
        key: typing.Optional[str] = None,
    ) -> None:
        self.type_id = type_id
        self.id = id
        self.key = key
        super().__init__()

    def __repr__(self) -> str:
        return "ResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class ScopedPrice(_BaseType):
    #: :class:`str`
    id: str
    #: :class:`commercetools.types.TypedMoney`
    value: "TypedMoney"
    #: :class:`commercetools.types.TypedMoney` `(Named` ``currentValue`` `in Commercetools)`
    current_value: "TypedMoney"
    #: Optional :class:`str`
    country: typing.Optional["str"]
    #: Optional :class:`commercetools.types.CustomerGroupReference` `(Named` ``customerGroup`` `in Commercetools)`
    customer_group: typing.Optional["CustomerGroupReference"]
    #: Optional :class:`commercetools.types.ChannelReference`
    channel: typing.Optional["ChannelReference"]
    #: Optional :class:`datetime.datetime` `(Named` ``validFrom`` `in Commercetools)`
    valid_from: typing.Optional[datetime.datetime]
    #: Optional :class:`datetime.datetime` `(Named` ``validUntil`` `in Commercetools)`
    valid_until: typing.Optional[datetime.datetime]
    #: Optional :class:`commercetools.types.DiscountedPrice`
    discounted: typing.Optional["DiscountedPrice"]
    #: Optional :class:`commercetools.types.CustomFields`
    custom: typing.Optional["CustomFields"]

    def __init__(
        self,
        *,
        id: str,
        value: "TypedMoney",
        current_value: "TypedMoney",
        country: typing.Optional["str"] = None,
        customer_group: typing.Optional["CustomerGroupReference"] = None,
        channel: typing.Optional["ChannelReference"] = None,
        valid_from: typing.Optional[datetime.datetime] = None,
        valid_until: typing.Optional[datetime.datetime] = None,
        discounted: typing.Optional["DiscountedPrice"] = None,
        custom: typing.Optional["CustomFields"] = None,
    ) -> None:
        self.id = id
        self.value = value
        self.current_value = current_value
        self.country = country
        self.customer_group = customer_group
        self.channel = channel
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.discounted = discounted
        self.custom = custom
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ScopedPrice(id=%r, value=%r, current_value=%r, country=%r, customer_group=%r, channel=%r, valid_from=%r, valid_until=%r, discounted=%r, custom=%r)"
            % (
                self.id,
                self.value,
                self.current_value,
                self.country,
                self.customer_group,
                self.channel,
                self.valid_from,
                self.valid_until,
                self.discounted,
                self.custom,
            )
        )


class TypedMoney(_BaseType):
    #: :class:`commercetools.types.MoneyType`
    type: "MoneyType"
    #: :class:`int` `(Named` ``fractionDigits`` `in Commercetools)`
    fraction_digits: int
    #: :class:`int` `(Named` ``centAmount`` `in Commercetools)`
    cent_amount: int
    #: :class:`str` `(Named` ``currencyCode`` `in Commercetools)`
    currency_code: "str"

    def __init__(
        self,
        *,
        type: "MoneyType",
        fraction_digits: int,
        cent_amount: int,
        currency_code: "str",
    ) -> None:
        self.type = type
        self.fraction_digits = fraction_digits
        self.cent_amount = cent_amount
        self.currency_code = currency_code
        super().__init__()

    def __repr__(self) -> str:
        return (
            "TypedMoney(type=%r, fraction_digits=%r, cent_amount=%r, currency_code=%r)"
            % (self.type, self.fraction_digits, self.cent_amount, self.currency_code)
        )


class Update(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "Update(version=%r, actions=%r)" % (self.version, self.actions)


class UpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "UpdateAction(action=%r)" % (self.action,)


class CentPrecisionMoney(TypedMoney):
    def __init__(
        self, *, fraction_digits: int, cent_amount: int, currency_code: "str"
    ) -> None:
        super().__init__(
            type=MoneyType.CENT_PRECISION,
            fraction_digits=fraction_digits,
            cent_amount=cent_amount,
            currency_code=currency_code,
        )

    def __repr__(self) -> str:
        return (
            "CentPrecisionMoney(type=%r, fraction_digits=%r, cent_amount=%r, currency_code=%r)"
            % (self.type, self.fraction_digits, self.cent_amount, self.currency_code)
        )


class CreatedBy(ClientLogging):
    def __init__(
        self,
        *,
        client_id: typing.Optional[str] = None,
        external_user_id: typing.Optional[str] = None,
        customer: typing.Optional["CustomerReference"] = None,
        anonymous_id: typing.Optional[str] = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            external_user_id=external_user_id,
            customer=customer,
            anonymous_id=anonymous_id,
        )

    def __repr__(self) -> str:
        return (
            "CreatedBy(client_id=%r, external_user_id=%r, customer=%r, anonymous_id=%r)"
            % (self.client_id, self.external_user_id, self.customer, self.anonymous_id)
        )


class GeoJsonPoint(GeoJson):
    #: :class:`list`
    coordinates: list

    def __init__(self, *, coordinates: list) -> None:
        self.coordinates = coordinates
        super().__init__(type="Point")

    def __repr__(self) -> str:
        return "GeoJsonPoint(type=%r, coordinates=%r)" % (self.type, self.coordinates)


class HighPrecisionMoney(TypedMoney):
    #: :class:`int` `(Named` ``preciseAmount`` `in Commercetools)`
    precise_amount: int

    def __init__(
        self,
        *,
        fraction_digits: int,
        cent_amount: int,
        currency_code: "str",
        precise_amount: int,
    ) -> None:
        self.precise_amount = precise_amount
        super().__init__(
            type=MoneyType.HIGH_PRECISION,
            fraction_digits=fraction_digits,
            cent_amount=cent_amount,
            currency_code=currency_code,
        )

    def __repr__(self) -> str:
        return (
            "HighPrecisionMoney(type=%r, fraction_digits=%r, cent_amount=%r, currency_code=%r, precise_amount=%r)"
            % (
                self.type,
                self.fraction_digits,
                self.cent_amount,
                self.currency_code,
                self.precise_amount,
            )
        )


class LastModifiedBy(ClientLogging):
    def __init__(
        self,
        *,
        client_id: typing.Optional[str] = None,
        external_user_id: typing.Optional[str] = None,
        customer: typing.Optional["CustomerReference"] = None,
        anonymous_id: typing.Optional[str] = None,
    ) -> None:
        super().__init__(
            client_id=client_id,
            external_user_id=external_user_id,
            customer=customer,
            anonymous_id=anonymous_id,
        )

    def __repr__(self) -> str:
        return (
            "LastModifiedBy(client_id=%r, external_user_id=%r, customer=%r, anonymous_id=%r)"
            % (self.client_id, self.external_user_id, self.customer, self.anonymous_id)
        )


class TypedMoneyDraft(Money):
    #: :class:`commercetools.types.MoneyType`
    type: "MoneyType"

    def __init__(
        self, *, cent_amount: int, currency_code: "str", type: "MoneyType"
    ) -> None:
        self.type = type
        super().__init__(cent_amount=cent_amount, currency_code=currency_code)

    def __repr__(self) -> str:
        return "TypedMoneyDraft(cent_amount=%r, currency_code=%r, type=%r)" % (
            self.cent_amount,
            self.currency_code,
            self.type,
        )


class CentPrecisionMoneyDraft(TypedMoneyDraft):
    def __init__(self, *, cent_amount: int, currency_code: "str") -> None:
        super().__init__(
            cent_amount=cent_amount,
            currency_code=currency_code,
            type=MoneyType.CENT_PRECISION,
        )

    def __repr__(self) -> str:
        return "CentPrecisionMoneyDraft(cent_amount=%r, currency_code=%r, type=%r)" % (
            self.cent_amount,
            self.currency_code,
            self.type,
        )


class HighPrecisionMoneyDraft(TypedMoneyDraft):
    #: :class:`int` `(Named` ``preciseAmount`` `in Commercetools)`
    precise_amount: int

    def __init__(
        self, *, cent_amount: int, currency_code: "str", precise_amount: int
    ) -> None:
        self.precise_amount = precise_amount
        super().__init__(
            cent_amount=cent_amount,
            currency_code=currency_code,
            type=MoneyType.HIGH_PRECISION,
        )

    def __repr__(self) -> str:
        return (
            "HighPrecisionMoneyDraft(cent_amount=%r, currency_code=%r, type=%r, precise_amount=%r)"
            % (self.cent_amount, self.currency_code, self.type, self.precise_amount)
        )



import pytest

from commercetools import CommercetoolsError, types


def test_correlation_id_is_set_in_exception(client):
    product = client.products.create(
        types.ProductDraft(
            key="test-product",
            name=types.LocalizedString(en=f"my-product"),
            slug=types.LocalizedString(en=f"my-product"),
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
        )
    )

    product = client.products.update_by_id(
        id=product.id,
        version=product.version,
        actions=[
            types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug2"))
        ],
    )

    # This should raise a version conflict error
    with pytest.raises(CommercetoolsError) as exc:
        client.products.update_by_id(
            id=product.id,
            version=1,
            actions=[
                types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug3"))
            ],
        )

    assert exc.value.correlation_id is not None



from commercetools import types


def test_get_by_id(client):
    customer_group = client.customer_groups.create(
        draft=types.CustomerGroupDraft(group_name="test customer group")
    )
    assert customer_group.id

    customer_group = client.customer_groups.get_by_id(customer_group.id)
    assert customer_group


def test_get_by_key(client):
    customer_group = client.customer_groups.create(
        draft=types.CustomerGroupDraft(
            group_name="test customer group", key="test-customer-group"
        )
    )
    assert customer_group.key

    customer_group = client.customer_groups.get_by_key(customer_group.key)
    assert customer_group


def test_query(client):
    client.customer_groups.create(
        draft=types.CustomerGroupDraft(
            key="test-customer-group-1", group_name="group-1"
        )
    )
    client.customer_groups.create(
        draft=types.CustomerGroupDraft(
            key="test-customer-group-2", group_name="group-2"
        )
    )

    result = client.customer_groups.query(sort="id asc", limit=10)
    assert len(result.results) == 2
    assert result.total == 2

    result = client.customer_groups.query(sort=["id asc", "name asc"], limit=1)
    assert len(result.results) == 1
    assert result.total == 2


def test_delete_by_id(client):
    customer_group = client.customer_groups.create(
        draft=types.CustomerGroupDraft(
            group_name="test customer group", key="test-customer-group"
        )
    )
    assert customer_group.id
    assert client.customer_groups.delete_by_id(
        customer_group.id, version=customer_group.version
    )


def test_delete_by_key(client):
    customer_group = client.customer_groups.create(
        draft=types.CustomerGroupDraft(
            group_name="test customer group", key="test-customer-group"
        )
    )
    assert customer_group.key
    assert client.customer_groups.delete_by_key(
        customer_group.key, version=customer_group.version
    )



import time
import uuid

import pytest

from commercetools import CommercetoolsError, types
from commercetools.client import Client


@pytest.fixture()
def client_environment_settings(monkeypatch):
    monkeypatch.setenv("CTP_PROJECT_KEY", "project_key")
    monkeypatch.setenv("CTP_CLIENT_ID", "client_id")
    monkeypatch.setenv("CTP_CLIENT_SECRET", "client_secret")
    monkeypatch.setenv("CTP_CLIENT_SCOPES", "some_scope")
    monkeypatch.setenv("CTP_API_URL", "https://api.shere.io")
    monkeypatch.setenv("CTP_AUTH_URL", "https://auth.sphere.io")


def test_client_with_environment_settings_is_setup(
    client_environment_settings, commercetools_api
):
    Client()
    assert len(commercetools_api.requests_mock.request_history) == 1


def test_auto_refresh(commercetools_api):
    commercetools_api.auth.set_expire_time(1)

    client = Client(
        client_id="unittest",
        client_secret="mysecret",
        project_key="test",
        url="https://api.sphere.io",
        token_url="https://auth.sphere.io/oauth/token",
    )
    client.products.query()
    time.sleep(1)
    client.products.query()
    client.products.query()

    auth_headers = set()
    for request in commercetools_api.requests_mock.request_history:
        if request.url.startswith("https://api.sphere.io/"):
            auth_headers.add(request.headers["Authorization"])

    assert len(auth_headers) == 2
    assert len(commercetools_api.auth.model.tokens) == 2


def test_cache_token(commercetools_api):
    Client(
        client_id="unittest",
        client_secret="none",
        project_key="test",
        url="https://api.sphere.io",
        token_url="https://auth.sphere.io/oauth/token",
    )
    assert len(commercetools_api.requests_mock.request_history) == 1

    Client(
        client_id="unittest",
        client_secret="none",
        project_key="test",
        url="https://api.sphere.io",
        token_url="https://auth.sphere.io/oauth/token",
    )
    assert len(commercetools_api.requests_mock.request_history) == 1


def test_resource_update_conflict(client):
    """Test the return value of the update methods.

    It doesn't test the actual update itself.
    TODO: See if this is worth testing since we're using a mocking backend
    """
    product = client.products.create(
        types.ProductDraft(
            key="test-product",
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
            name={"en": "my-product"},
            slug={"en": "foo-bar"},
        )
    )

    assert product.version == 1
    assert uuid.UUID(product.id)
    assert product.key == "test-product"

    product = client.products.update_by_id(
        id=product.id,
        version=product.version,
        actions=[
            types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug2"))
        ],
    )
    assert product.key == "test-product"
    assert product.version == 2

    # This should raise a version conflict error
    with pytest.raises(CommercetoolsError) as exc:
        product = client.products.update_by_id(
            id=product.id,
            version=1,
            actions=[
                types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug3"))
            ],
        )
    assert exc.value.response.status_code == 409
    assert exc.value.response.errors[0].current_version == 2

    # Force it
    product = client.products.update_by_id(
        id=product.id,
        version=1,
        actions=[
            types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug2"))
        ],
        force_update=True,
    )


def test_resource_delete_conflict(client):
    """Test the return value of the update methods.

    It doesn't test the actual update itself.
    TODO: See if this is worth testing since we're using a mocking backend
    """
    product = client.products.create(
        types.ProductDraft(
            key="test-product",
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
            name={"en": "my-product"},
            slug={"en": "foo-bar"},
        )
    )

    product = client.products.update_by_id(
        id=product.id,
        version=product.version,
        actions=[
            types.ProductChangeSlugAction(slug=types.LocalizedString(nl="nl-slug2"))
        ],
    )
    assert product.version == 2

    # This should raise a version conflict error
    with pytest.raises(CommercetoolsError) as exc:
        product = client.products.delete_by_id(id=product.id, version=1)

    assert exc.value.response.status_code == 409
    assert exc.value.response.errors[0].current_version == 2

    # Force it
    client.products.delete_by_id(id=product.id, version=1, force_delete=True)



from commercetools import types


def test_get_by_id(client):
    product = client.products.create(
        types.ProductDraft(
            master_variant=types.ProductVariantDraft(sku="123"),
            publish=True,
            name=types.LocalizedString(nl="Test product"),
            slug=types.LocalizedString(en=f"my-product"),
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
        )
    )

    shopping_list = client.shopping_lists.create(
        draft=types.ShoppingListDraft(
            name=types.LocalizedString({"nl": "Verlanglijstje"}),
            description=types.LocalizedString({"nl": "Verlanglijstje van LabD"}),
            line_items=[
                types.ShoppingListLineItemDraft(product_id=product.id, quantity=1)
            ],
        )
    )
    assert shopping_list.id

    shopping_list = client.shopping_lists.get_by_id(shopping_list.id)
    assert shopping_list.name["nl"] == "Verlanglijstje"
    assert shopping_list.description["nl"] == "Verlanglijstje van LabD"
    assert shopping_list.line_items[0].product_id == product.id
    assert shopping_list.line_items[0].quantity == 1


def test_get_by_key(client):
    product = client.products.create(
        types.ProductDraft(
            master_variant=types.ProductVariantDraft(sku="123"),
            publish=True,
            name=types.LocalizedString(nl="Test product"),
            slug=types.LocalizedString(en="my-product"),
            product_type=types.ProductTypeResourceIdentifier(key="dummy"),
        )
    )

    variant = product.master_data.current.master_variant
    shopping_list = client.shopping_lists.create(
        draft=types.ShoppingListDraft(
            key="test-shopping-list",
            name=types.LocalizedString({"nl": "Verlanglijstje"}),
            description=types.LocalizedString({"nl": "Verlanglijstje van LabD"}),
            line_items=[types.ShoppingListLineItemDraft(sku=variant.sku, quantity=1)],
        )
    )
    assert shopping_list.key

    shopping_list = client.shopping_lists.get_by_key("test-shopping-list")
    assert shopping_list.name["nl"] == "Verlanglijstje"
    assert shopping_list.description["nl"] == "Verlanglijstje van LabD"
    assert shopping_list.line_items[0].variant.sku == "123"
    assert shopping_list.line_items[0].quantity == 1


def test_query(client):
    shopping_list_draft = types.ShoppingListDraft(
        key="test-shopping-list",
        name=types.LocalizedString({"nl": "Verlanglijstje"}),
        description=types.LocalizedString({"nl": "Verlanglijstje van LabD"}),
    )

    client.shopping_lists.create(draft=shopping_list_draft)

    # Update the key and create another one.
    shopping_list_draft.key = "test-shopping-list2"
    client.shopping_lists.create(draft=shopping_list_draft)

    result = client.shopping_lists.query(sort="id asc", limit=10)
    assert len(result.results) == 2
    assert result.total == 2

    result = client.shopping_lists.query(sort=["id asc", "name asc"], limit=1)
    assert len(result.results) == 1
    assert result.total == 2


def test_delete_by_id(client):
    shopping_list = client.shopping_lists.create(
        draft=types.ShoppingListDraft(
            key="test-shopping-list",
            name=types.LocalizedString({"nl": "Verlanglijstje"}),
            description=types.LocalizedString({"nl": "Verlanglijstje van LabD"}),
        )
    )
    assert shopping_list.id

    shopping_list = client.shopping_lists.delete_by_id(
        shopping_list.id, version=shopping_list.version
    )


def test_delete_by_key(client):
    shopping_list = client.shopping_lists.create(
        draft=types.ShoppingListDraft(
            key="test-shopping-list",
            name=types.LocalizedString({"nl": "Verlanglijstje"}),
            description=types.LocalizedString({"nl": "Verlanglijstje van LabD"}),
        )
    )
    assert shopping_list.id

    shopping_list = client.shopping_lists.delete_by_key(
        shopping_list.key, version=shopping_list.version
    )



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._type import (
    FieldDefinitionSchema,
    TypeDraftSchema,
    TypePagedQueryResponseSchema,
    TypeSchema,
    TypeUpdateSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import update_attribute, update_attribute_add_item


class TypesModel(BaseModel):
    _primary_type_name = "type"
    _resource_schema = TypeSchema
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.TypeDraft, id: typing.Optional[str] = None
    ) -> types.Type:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.Type(
            id=str(object_id),
            version=1,
            name=draft.name,
            description=draft.description,
            key=draft.key,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            resource_type_ids=draft.resource_type_ids,
            field_definitions=draft.field_definitions,
        )


class TypesBackend(ServiceBackend):
    service_path = "types"
    model_class = TypesModel
    _schema_draft = TypeDraftSchema
    _schema_update = TypeUpdateSchema
    _schema_query_response = TypePagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
        ]

    # Fixme: use decorator for this
    _actions = {
        "addFieldDefinition": update_attribute_add_item(
            "fieldDefinitions", "field_definition", FieldDefinitionSchema
        ),
        "setDescription": update_attribute("description", "description"),
    }



import pytest

from commercetools import paginators, types


def create_products(client):
    for i in range(100):
        client.products.create(
            types.ProductDraft(
                key=f"test-product-{i}",
                product_type=types.ProductTypeResourceIdentifier(key="dummy"),
                name=types.LocalizedString(en=f"my-product-{i}"),
                slug=types.LocalizedString(en=f"my-product-{i}"),
            )
        )


def test_page_paginator(client):
    create_products(client)
    paginator = paginators.Paginator(client.products.query, sort=["id asc", "name asc"])

    items = []
    for product in paginator:
        items.append(product)
    assert len(items) == 100


def test_page_paginator_slice_start(client):
    create_products(client)
    paginator = paginators.Paginator(client.products.query, sort=["id asc", "name asc"])

    items = []
    for product in paginator[20:]:
        items.append(product)
        if len(items) > 80:
            assert False
    assert len(items) == 80


def test_page_paginator_slice_stop(client):
    create_products(client)
    paginator = paginators.Paginator(client.products.query, sort=["id asc", "name asc"])

    items = []
    for product in paginator[:-20]:
        items.append(product)
        if len(items) > 80:
            assert False
    assert len(items) == 80


def test_page_paginator_slice_start_stop(client):
    create_products(client)
    paginator = paginators.Paginator(client.products.query, sort=["id asc", "name asc"])

    items = []
    for product in paginator[20:-20]:
        items.append(product)
        if len(items) > 60:
            assert False
    assert len(items) == 60


def test_cursor_paginator(client):
    create_products(client)
    paginator = paginators.CursorPaginator(
        client.products.query, sort=["id asc", "name asc"]
    )

    items = []
    for product in paginator:
        items.append(product)
    assert len(items) == 100


def test_cursor_paginator_slice_start(client):
    create_products(client)
    paginator = paginators.CursorPaginator(
        client.products.query, sort=["id asc", "name asc"]
    )

    items = []
    with pytest.raises(ValueError):
        for product in paginator[10:]:
            items.append(product)


def test_cursor_paginator_slice_stop(client):
    create_products(client)
    paginator = paginators.CursorPaginator(
        client.products.query, sort=["id asc", "name asc"]
    )

    items = []
    for product in paginator[:90]:
        items.append(product)

    assert len(items) == 90



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._custom_object import (
    CustomObjectDraftSchema,
    CustomObjectPagedQueryResponseSchema,
    CustomObjectSchema,
)
from commercetools._schemas._error import ErrorResponseSchema
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import create_commercetools_response


class CustomObjectsModel(BaseModel):
    _primary_type_name = "custom-object"
    _resource_schema = CustomObjectSchema

    def _generate_key(self, obj):
        return obj.container, obj.key

    def _create_from_draft(
        self, obj: types.CustomObjectDraft, id: typing.Optional[str] = None
    ) -> types.CustomObject:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.CustomObject(
            id=str(object_id),
            version=1,
            key=obj.key,
            container=obj.container,
            value=obj.value,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
        )


class CustomObjectsBackend(ServiceBackend):
    service_path = "custom-objects"
    model_class = CustomObjectsModel
    _schema_draft = CustomObjectDraftSchema
    _schema_query_response = CustomObjectPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<container>[^/]+)/(?P<key>[^/]+)$", "GET", self.get_by_container_key),
            ("^(?P<container>[^/]+)$", "GET", self.query_by_container),
        ]

    def query_by_container(self, request, container: str):
        # container is not a valid predicate filter, but we can abuse it internally.
        if "where" not in request.qs:
            request.qs["where"] = f'container = "{container}"'
        else:
            request.qs["where"] = f'container = "{container}" AND {request.qs["where"]}'

        return self.query(request)

    def get_by_container_key(self, request, container: str, key: str):
        item = next(
            (
                obj
                for obj in self.model.objects.values()
                if obj["container"] == container and obj["key"] == key
            ),
            None,
        )
        if item:
            return create_commercetools_response(request, json=item)
        else:
            return create_commercetools_response(request, status_code=404)



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._shipping_method import ShippingRateTierType

if typing.TYPE_CHECKING:
    from ._message import MessageConfiguration, MessageConfigurationDraft
__all__ = [
    "CartClassificationType",
    "CartScoreType",
    "CartValueType",
    "CartsConfiguration",
    "ExternalOAuth",
    "Project",
    "ProjectChangeCountriesAction",
    "ProjectChangeCountryTaxRateFallbackEnabledAction",
    "ProjectChangeCurrenciesAction",
    "ProjectChangeLanguagesAction",
    "ProjectChangeMessagesConfigurationAction",
    "ProjectChangeMessagesEnabledAction",
    "ProjectChangeNameAction",
    "ProjectSetExternalOAuthAction",
    "ProjectSetShippingRateInputTypeAction",
    "ProjectUpdate",
    "ProjectUpdateAction",
    "ShippingRateInputType",
]


class CartsConfiguration(_BaseType):
    #: Optional :class:`bool` `(Named` ``countryTaxRateFallbackEnabled`` `in Commercetools)`
    country_tax_rate_fallback_enabled: typing.Optional[bool]

    def __init__(
        self, *, country_tax_rate_fallback_enabled: typing.Optional[bool] = None
    ) -> None:
        self.country_tax_rate_fallback_enabled = country_tax_rate_fallback_enabled
        super().__init__()

    def __repr__(self) -> str:
        return "CartsConfiguration(country_tax_rate_fallback_enabled=%r)" % (
            self.country_tax_rate_fallback_enabled,
        )


class ExternalOAuth(_BaseType):
    #: :class:`str`
    url: str
    #: :class:`str` `(Named` ``authorizationHeader`` `in Commercetools)`
    authorization_header: str

    def __init__(self, *, url: str, authorization_header: str) -> None:
        self.url = url
        self.authorization_header = authorization_header
        super().__init__()

    def __repr__(self) -> str:
        return "ExternalOAuth(url=%r, authorization_header=%r)" % (
            self.url,
            self.authorization_header,
        )


class Project(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`str`
    key: str
    #: :class:`str`
    name: str
    #: List of :class:`str`
    countries: typing.List["str"]
    #: List of :class:`str`
    currencies: typing.List["str"]
    #: List of :class:`str`
    languages: typing.List["str"]
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: Optional :class:`str` `(Named` ``trialUntil`` `in Commercetools)`
    trial_until: typing.Optional[str]
    #: :class:`commercetools.types.MessageConfiguration`
    messages: "MessageConfiguration"
    #: Optional :class:`commercetools.types.ShippingRateInputType` `(Named` ``shippingRateInputType`` `in Commercetools)`
    shipping_rate_input_type: typing.Optional["ShippingRateInputType"]
    #: Optional :class:`commercetools.types.ExternalOAuth` `(Named` ``externalOAuth`` `in Commercetools)`
    external_oauth: typing.Optional["ExternalOAuth"]
    #: :class:`commercetools.types.CartsConfiguration`
    carts: "CartsConfiguration"

    def __init__(
        self,
        *,
        version: int,
        key: str,
        name: str,
        countries: typing.List["str"],
        currencies: typing.List["str"],
        languages: typing.List["str"],
        created_at: datetime.datetime,
        messages: "MessageConfiguration",
        carts: "CartsConfiguration",
        trial_until: typing.Optional[str] = None,
        shipping_rate_input_type: typing.Optional["ShippingRateInputType"] = None,
        external_oauth: typing.Optional["ExternalOAuth"] = None
    ) -> None:
        self.version = version
        self.key = key
        self.name = name
        self.countries = countries
        self.currencies = currencies
        self.languages = languages
        self.created_at = created_at
        self.trial_until = trial_until
        self.messages = messages
        self.shipping_rate_input_type = shipping_rate_input_type
        self.external_oauth = external_oauth
        self.carts = carts
        super().__init__()

    def __repr__(self) -> str:
        return (
            "Project(version=%r, key=%r, name=%r, countries=%r, currencies=%r, languages=%r, created_at=%r, trial_until=%r, messages=%r, shipping_rate_input_type=%r, external_oauth=%r, carts=%r)"
            % (
                self.version,
                self.key,
                self.name,
                self.countries,
                self.currencies,
                self.languages,
                self.created_at,
                self.trial_until,
                self.messages,
                self.shipping_rate_input_type,
                self.external_oauth,
                self.carts,
            )
        )


class ProjectUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "ProjectUpdate(version=%r, actions=%r)" % (self.version, self.actions)


class ProjectUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "ProjectUpdateAction(action=%r)" % (self.action,)


class ShippingRateInputType(_BaseType):
    #: :class:`commercetools.types.ShippingRateTierType`
    type: "ShippingRateTierType"

    def __init__(self, *, type: "ShippingRateTierType") -> None:
        self.type = type
        super().__init__()

    def __repr__(self) -> str:
        return "ShippingRateInputType(type=%r)" % (self.type,)


class CartClassificationType(ShippingRateInputType):
    #: :class:`list`
    values: list

    def __init__(self, *, values: list) -> None:
        self.values = values
        super().__init__(type=ShippingRateTierType.CART_CLASSIFICATION)

    def __repr__(self) -> str:
        return "CartClassificationType(type=%r, values=%r)" % (self.type, self.values)


class CartScoreType(ShippingRateInputType):
    def __init__(self) -> None:
        super().__init__(type=ShippingRateTierType.CART_SCORE)

    def __repr__(self) -> str:
        return "CartScoreType(type=%r)" % (self.type,)


class CartValueType(ShippingRateInputType):
    def __init__(self) -> None:
        super().__init__(type=ShippingRateTierType.CART_VALUE)

    def __repr__(self) -> str:
        return "CartValueType(type=%r)" % (self.type,)


class ProjectChangeCountriesAction(ProjectUpdateAction):
    #: List of :class:`str`
    countries: typing.List["str"]

    def __init__(self, *, countries: typing.List["str"]) -> None:
        self.countries = countries
        super().__init__(action="changeCountries")

    def __repr__(self) -> str:
        return "ProjectChangeCountriesAction(action=%r, countries=%r)" % (
            self.action,
            self.countries,
        )


class ProjectChangeCountryTaxRateFallbackEnabledAction(ProjectUpdateAction):
    #: :class:`bool` `(Named` ``countryTaxRateFallbackEnabled`` `in Commercetools)`
    country_tax_rate_fallback_enabled: bool

    def __init__(self, *, country_tax_rate_fallback_enabled: bool) -> None:
        self.country_tax_rate_fallback_enabled = country_tax_rate_fallback_enabled
        super().__init__(action="changeCountryTaxRateFallbackEnabled")

    def __repr__(self) -> str:
        return (
            "ProjectChangeCountryTaxRateFallbackEnabledAction(action=%r, country_tax_rate_fallback_enabled=%r)"
            % (self.action, self.country_tax_rate_fallback_enabled)
        )


class ProjectChangeCurrenciesAction(ProjectUpdateAction):
    #: List of :class:`str`
    currencies: typing.List["str"]

    def __init__(self, *, currencies: typing.List["str"]) -> None:
        self.currencies = currencies
        super().__init__(action="changeCurrencies")

    def __repr__(self) -> str:
        return "ProjectChangeCurrenciesAction(action=%r, currencies=%r)" % (
            self.action,
            self.currencies,
        )


class ProjectChangeLanguagesAction(ProjectUpdateAction):
    #: List of :class:`str`
    languages: typing.List["str"]

    def __init__(self, *, languages: typing.List["str"]) -> None:
        self.languages = languages
        super().__init__(action="changeLanguages")

    def __repr__(self) -> str:
        return "ProjectChangeLanguagesAction(action=%r, languages=%r)" % (
            self.action,
            self.languages,
        )


class ProjectChangeMessagesConfigurationAction(ProjectUpdateAction):
    #: :class:`commercetools.types.MessageConfigurationDraft` `(Named` ``messagesConfiguration`` `in Commercetools)`
    messages_configuration: "MessageConfigurationDraft"

    def __init__(self, *, messages_configuration: "MessageConfigurationDraft") -> None:
        self.messages_configuration = messages_configuration
        super().__init__(action="changeMessagesConfiguration")

    def __repr__(self) -> str:
        return (
            "ProjectChangeMessagesConfigurationAction(action=%r, messages_configuration=%r)"
            % (self.action, self.messages_configuration)
        )


class ProjectChangeMessagesEnabledAction(ProjectUpdateAction):
    #: :class:`bool` `(Named` ``messagesEnabled`` `in Commercetools)`
    messages_enabled: bool

    def __init__(self, *, messages_enabled: bool) -> None:
        self.messages_enabled = messages_enabled
        super().__init__(action="changeMessagesEnabled")

    def __repr__(self) -> str:
        return "ProjectChangeMessagesEnabledAction(action=%r, messages_enabled=%r)" % (
            self.action,
            self.messages_enabled,
        )


class ProjectChangeNameAction(ProjectUpdateAction):
    #: :class:`str`
    name: str

    def __init__(self, *, name: str) -> None:
        self.name = name
        super().__init__(action="changeName")

    def __repr__(self) -> str:
        return "ProjectChangeNameAction(action=%r, name=%r)" % (self.action, self.name)


class ProjectSetExternalOAuthAction(ProjectUpdateAction):
    #: Optional :class:`commercetools.types.ExternalOAuth` `(Named` ``externalOAuth`` `in Commercetools)`
    external_oauth: typing.Optional["ExternalOAuth"]

    def __init__(
        self, *, external_oauth: typing.Optional["ExternalOAuth"] = None
    ) -> None:
        self.external_oauth = external_oauth
        super().__init__(action="setExternalOAuth")

    def __repr__(self) -> str:
        return "ProjectSetExternalOAuthAction(action=%r, external_oauth=%r)" % (
            self.action,
            self.external_oauth,
        )


class ProjectSetShippingRateInputTypeAction(ProjectUpdateAction):
    #: Optional :class:`commercetools.types.ShippingRateInputType` `(Named` ``shippingRateInputType`` `in Commercetools)`
    shipping_rate_input_type: typing.Optional["ShippingRateInputType"]

    def __init__(
        self,
        *,
        shipping_rate_input_type: typing.Optional["ShippingRateInputType"] = None
    ) -> None:
        self.shipping_rate_input_type = shipping_rate_input_type
        super().__init__(action="setShippingRateInputType")

    def __repr__(self) -> str:
        return (
            "ProjectSetShippingRateInputTypeAction(action=%r, shipping_rate_input_type=%r)"
            % (self.action, self.shipping_rate_input_type)
        )



# DO NOT EDIT! This file is automatically generated
import datetime
import typing

from commercetools.types._abstract import _BaseType
from commercetools.types._common import (
    BaseResource,
    Reference,
    ReferenceTypeId,
    ResourceIdentifier,
)

if typing.TYPE_CHECKING:
    from ._common import CreatedBy, LastModifiedBy
__all__ = [
    "Location",
    "Zone",
    "ZoneAddLocationAction",
    "ZoneChangeNameAction",
    "ZoneDraft",
    "ZonePagedQueryResponse",
    "ZoneReference",
    "ZoneRemoveLocationAction",
    "ZoneResourceIdentifier",
    "ZoneSetDescriptionAction",
    "ZoneSetKeyAction",
    "ZoneUpdate",
    "ZoneUpdateAction",
]


class Location(_BaseType):
    #: :class:`str`
    country: "str"
    #: Optional :class:`str`
    state: typing.Optional[str]

    def __init__(self, *, country: "str", state: typing.Optional[str] = None) -> None:
        self.country = country
        self.state = state
        super().__init__()

    def __repr__(self) -> str:
        return "Location(country=%r, state=%r)" % (self.country, self.state)


class Zone(BaseResource):
    #: :class:`str`
    id: str
    #: :class:`int`
    version: int
    #: :class:`datetime.datetime` `(Named` ``createdAt`` `in Commercetools)`
    created_at: datetime.datetime
    #: :class:`datetime.datetime` `(Named` ``lastModifiedAt`` `in Commercetools)`
    last_modified_at: datetime.datetime
    #: Optional :class:`commercetools.types.LastModifiedBy` `(Named` ``lastModifiedBy`` `in Commercetools)`
    last_modified_by: typing.Optional["LastModifiedBy"]
    #: Optional :class:`commercetools.types.CreatedBy` `(Named` ``createdBy`` `in Commercetools)`
    created_by: typing.Optional["CreatedBy"]
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: :class:`str`
    name: str
    #: Optional :class:`str`
    description: typing.Optional[str]
    #: List of :class:`commercetools.types.Location`
    locations: typing.List["Location"]

    def __init__(
        self,
        *,
        id: str,
        version: int,
        created_at: datetime.datetime,
        last_modified_at: datetime.datetime,
        name: str,
        locations: typing.List["Location"],
        last_modified_by: typing.Optional["LastModifiedBy"] = None,
        created_by: typing.Optional["CreatedBy"] = None,
        key: typing.Optional[str] = None,
        description: typing.Optional[str] = None
    ) -> None:
        self.id = id
        self.version = version
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        self.last_modified_by = last_modified_by
        self.created_by = created_by
        self.key = key
        self.name = name
        self.description = description
        self.locations = locations
        super().__init__(
            id=id,
            version=version,
            created_at=created_at,
            last_modified_at=last_modified_at,
        )

    def __repr__(self) -> str:
        return (
            "Zone(id=%r, version=%r, created_at=%r, last_modified_at=%r, last_modified_by=%r, created_by=%r, key=%r, name=%r, description=%r, locations=%r)"
            % (
                self.id,
                self.version,
                self.created_at,
                self.last_modified_at,
                self.last_modified_by,
                self.created_by,
                self.key,
                self.name,
                self.description,
                self.locations,
            )
        )


class ZoneDraft(_BaseType):
    #: Optional :class:`str`
    key: typing.Optional[str]
    #: :class:`str`
    name: str
    #: Optional :class:`str`
    description: typing.Optional[str]
    #: List of :class:`commercetools.types.Location`
    locations: typing.List["Location"]

    def __init__(
        self,
        *,
        name: str,
        locations: typing.List["Location"],
        key: typing.Optional[str] = None,
        description: typing.Optional[str] = None
    ) -> None:
        self.key = key
        self.name = name
        self.description = description
        self.locations = locations
        super().__init__()

    def __repr__(self) -> str:
        return "ZoneDraft(key=%r, name=%r, description=%r, locations=%r)" % (
            self.key,
            self.name,
            self.description,
            self.locations,
        )


class ZonePagedQueryResponse(_BaseType):
    #: :class:`int`
    limit: int
    #: :class:`int`
    count: int
    #: Optional :class:`int`
    total: typing.Optional[int]
    #: :class:`int`
    offset: int
    #: List of :class:`commercetools.types.Zone`
    results: typing.Sequence["Zone"]

    def __init__(
        self,
        *,
        limit: int,
        count: int,
        offset: int,
        results: typing.Sequence["Zone"],
        total: typing.Optional[int] = None
    ) -> None:
        self.limit = limit
        self.count = count
        self.total = total
        self.offset = offset
        self.results = results
        super().__init__()

    def __repr__(self) -> str:
        return (
            "ZonePagedQueryResponse(limit=%r, count=%r, total=%r, offset=%r, results=%r)"
            % (self.limit, self.count, self.total, self.offset, self.results)
        )


class ZoneReference(Reference):
    #: Optional :class:`commercetools.types.Zone`
    obj: typing.Optional["Zone"]

    def __init__(self, *, id: str, obj: typing.Optional["Zone"] = None) -> None:
        self.obj = obj
        super().__init__(type_id=ReferenceTypeId.ZONE, id=id)

    def __repr__(self) -> str:
        return "ZoneReference(type_id=%r, id=%r, obj=%r)" % (
            self.type_id,
            self.id,
            self.obj,
        )


class ZoneResourceIdentifier(ResourceIdentifier):
    def __init__(
        self, *, id: typing.Optional[str] = None, key: typing.Optional[str] = None
    ) -> None:
        super().__init__(type_id=ReferenceTypeId.ZONE, id=id, key=key)

    def __repr__(self) -> str:
        return "ZoneResourceIdentifier(type_id=%r, id=%r, key=%r)" % (
            self.type_id,
            self.id,
            self.key,
        )


class ZoneUpdate(_BaseType):
    #: :class:`int`
    version: int
    #: :class:`list`
    actions: list

    def __init__(self, *, version: int, actions: list) -> None:
        self.version = version
        self.actions = actions
        super().__init__()

    def __repr__(self) -> str:
        return "ZoneUpdate(version=%r, actions=%r)" % (self.version, self.actions)


class ZoneUpdateAction(_BaseType):
    #: :class:`str`
    action: str

    def __init__(self, *, action: str) -> None:
        self.action = action
        super().__init__()

    def __repr__(self) -> str:
        return "ZoneUpdateAction(action=%r)" % (self.action,)


class ZoneAddLocationAction(ZoneUpdateAction):
    #: :class:`commercetools.types.Location`
    location: "Location"

    def __init__(self, *, location: "Location") -> None:
        self.location = location
        super().__init__(action="addLocation")

    def __repr__(self) -> str:
        return "ZoneAddLocationAction(action=%r, location=%r)" % (
            self.action,
            self.location,
        )


class ZoneChangeNameAction(ZoneUpdateAction):
    #: :class:`str`
    name: str

    def __init__(self, *, name: str) -> None:
        self.name = name
        super().__init__(action="changeName")

    def __repr__(self) -> str:
        return "ZoneChangeNameAction(action=%r, name=%r)" % (self.action, self.name)


class ZoneRemoveLocationAction(ZoneUpdateAction):
    #: :class:`commercetools.types.Location`
    location: "Location"

    def __init__(self, *, location: "Location") -> None:
        self.location = location
        super().__init__(action="removeLocation")

    def __repr__(self) -> str:
        return "ZoneRemoveLocationAction(action=%r, location=%r)" % (
            self.action,
            self.location,
        )


class ZoneSetDescriptionAction(ZoneUpdateAction):
    #: Optional :class:`str`
    description: typing.Optional[str]

    def __init__(self, *, description: typing.Optional[str] = None) -> None:
        self.description = description
        super().__init__(action="setDescription")

    def __repr__(self) -> str:
        return "ZoneSetDescriptionAction(action=%r, description=%r)" % (
            self.action,
            self.description,
        )


class ZoneSetKeyAction(ZoneUpdateAction):
    #: Optional :class:`str`
    key: typing.Optional[str]

    def __init__(self, *, key: typing.Optional[str] = None) -> None:
        self.key = key
        super().__init__(action="setKey")

    def __repr__(self) -> str:
        return "ZoneSetKeyAction(action=%r, key=%r)" % (self.action, self.key)



# DO NOT EDIT! This file is automatically generated
import typing

if typing.TYPE_CHECKING:
    from commercetools.protocols import ClientProtocol

    from .api_clients import ApiClientService
    from .cart_discounts import CartDiscountService
    from .carts import CartService
    from .categories import CategoryService
    from .channels import ChannelService
    from .custom_objects import CustomObjectService
    from .customer_groups import CustomerGroupService
    from .customers import CustomerService
    from .discount_codes import DiscountCodeService
    from .extensions import ExtensionService
    from .graphqls import GraphqlService
    from .in_stores import In_StoreService
    from .inventory import InventoryEntryService
    from .login import LoginService
    from .me import MeService
    from .messages import MessageService
    from .orders import OrderService
    from .payments import PaymentService
    from .product_discounts import ProductDiscountService
    from .product_projections import ProductProjectionService
    from .product_types import ProductTypeService
    from .products import ProductService
    from .project import ProjectService
    from .reviews import ReviewService
    from .shipping_methods import ShippingMethodService
    from .shopping_lists import ShoppingListService
    from .states import StateService
    from .stores import StoreService
    from .subscriptions import SubscriptionService
    from .tax_categories import TaxCategoryService
    from .types import TypeService
    from .zones import ZoneService


class ServicesMixin:
    @property
    def categories(self: "ClientProtocol") -> "CategoryService":
        if "categories" not in self.__dict__:
            from .categories import CategoryService

            self.__dict__["categories"] = CategoryService(self)
        return self.__dict__["categories"]

    @property
    def carts(self: "ClientProtocol") -> "CartService":
        if "carts" not in self.__dict__:
            from .carts import CartService

            self.__dict__["carts"] = CartService(self)
        return self.__dict__["carts"]

    @property
    def cart_discounts(self: "ClientProtocol") -> "CartDiscountService":
        if "cart_discounts" not in self.__dict__:
            from .cart_discounts import CartDiscountService

            self.__dict__["cart_discounts"] = CartDiscountService(self)
        return self.__dict__["cart_discounts"]

    @property
    def channels(self: "ClientProtocol") -> "ChannelService":
        if "channels" not in self.__dict__:
            from .channels import ChannelService

            self.__dict__["channels"] = ChannelService(self)
        return self.__dict__["channels"]

    @property
    def customers(self: "ClientProtocol") -> "CustomerService":
        if "customers" not in self.__dict__:
            from .customers import CustomerService

            self.__dict__["customers"] = CustomerService(self)
        return self.__dict__["customers"]

    @property
    def customer_groups(self: "ClientProtocol") -> "CustomerGroupService":
        if "customer_groups" not in self.__dict__:
            from .customer_groups import CustomerGroupService

            self.__dict__["customer_groups"] = CustomerGroupService(self)
        return self.__dict__["customer_groups"]

    @property
    def custom_objects(self: "ClientProtocol") -> "CustomObjectService":
        if "custom_objects" not in self.__dict__:
            from .custom_objects import CustomObjectService

            self.__dict__["custom_objects"] = CustomObjectService(self)
        return self.__dict__["custom_objects"]

    @property
    def discount_codes(self: "ClientProtocol") -> "DiscountCodeService":
        if "discount_codes" not in self.__dict__:
            from .discount_codes import DiscountCodeService

            self.__dict__["discount_codes"] = DiscountCodeService(self)
        return self.__dict__["discount_codes"]

    @property
    def graphqls(self: "ClientProtocol") -> "GraphqlService":
        if "graphqls" not in self.__dict__:
            from .graphqls import GraphqlService

            self.__dict__["graphqls"] = GraphqlService(self)
        return self.__dict__["graphqls"]

    @property
    def inventory(self: "ClientProtocol") -> "InventoryEntryService":
        if "inventory" not in self.__dict__:
            from .inventory import InventoryEntryService

            self.__dict__["inventory"] = InventoryEntryService(self)
        return self.__dict__["inventory"]

    @property
    def login(self: "ClientProtocol") -> "LoginService":
        if "login" not in self.__dict__:
            from .login import LoginService

            self.__dict__["login"] = LoginService(self)
        return self.__dict__["login"]

    @property
    def messages(self: "ClientProtocol") -> "MessageService":
        if "messages" not in self.__dict__:
            from .messages import MessageService

            self.__dict__["messages"] = MessageService(self)
        return self.__dict__["messages"]

    @property
    def orders(self: "ClientProtocol") -> "OrderService":
        if "orders" not in self.__dict__:
            from .orders import OrderService

            self.__dict__["orders"] = OrderService(self)
        return self.__dict__["orders"]

    @property
    def payments(self: "ClientProtocol") -> "PaymentService":
        if "payments" not in self.__dict__:
            from .payments import PaymentService

            self.__dict__["payments"] = PaymentService(self)
        return self.__dict__["payments"]

    @property
    def products(self: "ClientProtocol") -> "ProductService":
        if "products" not in self.__dict__:
            from .products import ProductService

            self.__dict__["products"] = ProductService(self)
        return self.__dict__["products"]

    @property
    def product_discounts(self: "ClientProtocol") -> "ProductDiscountService":
        if "product_discounts" not in self.__dict__:
            from .product_discounts import ProductDiscountService

            self.__dict__["product_discounts"] = ProductDiscountService(self)
        return self.__dict__["product_discounts"]

    @property
    def product_projections(self: "ClientProtocol") -> "ProductProjectionService":
        if "product_projections" not in self.__dict__:
            from .product_projections import ProductProjectionService

            self.__dict__["product_projections"] = ProductProjectionService(self)
        return self.__dict__["product_projections"]

    @property
    def product_types(self: "ClientProtocol") -> "ProductTypeService":
        if "product_types" not in self.__dict__:
            from .product_types import ProductTypeService

            self.__dict__["product_types"] = ProductTypeService(self)
        return self.__dict__["product_types"]

    @property
    def reviews(self: "ClientProtocol") -> "ReviewService":
        if "reviews" not in self.__dict__:
            from .reviews import ReviewService

            self.__dict__["reviews"] = ReviewService(self)
        return self.__dict__["reviews"]

    @property
    def shipping_methods(self: "ClientProtocol") -> "ShippingMethodService":
        if "shipping_methods" not in self.__dict__:
            from .shipping_methods import ShippingMethodService

            self.__dict__["shipping_methods"] = ShippingMethodService(self)
        return self.__dict__["shipping_methods"]

    @property
    def shopping_lists(self: "ClientProtocol") -> "ShoppingListService":
        if "shopping_lists" not in self.__dict__:
            from .shopping_lists import ShoppingListService

            self.__dict__["shopping_lists"] = ShoppingListService(self)
        return self.__dict__["shopping_lists"]

    @property
    def states(self: "ClientProtocol") -> "StateService":
        if "states" not in self.__dict__:
            from .states import StateService

            self.__dict__["states"] = StateService(self)
        return self.__dict__["states"]

    @property
    def subscriptions(self: "ClientProtocol") -> "SubscriptionService":
        if "subscriptions" not in self.__dict__:
            from .subscriptions import SubscriptionService

            self.__dict__["subscriptions"] = SubscriptionService(self)
        return self.__dict__["subscriptions"]

    @property
    def tax_categories(self: "ClientProtocol") -> "TaxCategoryService":
        if "tax_categories" not in self.__dict__:
            from .tax_categories import TaxCategoryService

            self.__dict__["tax_categories"] = TaxCategoryService(self)
        return self.__dict__["tax_categories"]

    @property
    def types(self: "ClientProtocol") -> "TypeService":
        if "types" not in self.__dict__:
            from .types import TypeService

            self.__dict__["types"] = TypeService(self)
        return self.__dict__["types"]

    @property
    def zones(self: "ClientProtocol") -> "ZoneService":
        if "zones" not in self.__dict__:
            from .zones import ZoneService

            self.__dict__["zones"] = ZoneService(self)
        return self.__dict__["zones"]

    @property
    def me(self: "ClientProtocol") -> "MeService":
        if "me" not in self.__dict__:
            from .me import MeService

            self.__dict__["me"] = MeService(self)
        return self.__dict__["me"]

    @property
    def extensions(self: "ClientProtocol") -> "ExtensionService":
        if "extensions" not in self.__dict__:
            from .extensions import ExtensionService

            self.__dict__["extensions"] = ExtensionService(self)
        return self.__dict__["extensions"]

    @property
    def api_clients(self: "ClientProtocol") -> "ApiClientService":
        if "api_clients" not in self.__dict__:
            from .api_clients import ApiClientService

            self.__dict__["api_clients"] = ApiClientService(self)
        return self.__dict__["api_clients"]

    @property
    def stores(self: "ClientProtocol") -> "StoreService":
        if "stores" not in self.__dict__:
            from .stores import StoreService

            self.__dict__["stores"] = StoreService(self)
        return self.__dict__["stores"]

    @property
    def in_stores(self: "ClientProtocol") -> "In_StoreService":
        if "in_stores" not in self.__dict__:
            from .in_stores import In_StoreService

            self.__dict__["in_stores"] = In_StoreService(self)
        return self.__dict__["in_stores"]

    @property
    def project(self: "ClientProtocol") -> "ProjectService":
        if "project" not in self.__dict__:
            from .project import ProjectService

            self.__dict__["project"] = ProjectService(self)
        return self.__dict__["project"]
